# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0


class EnvironmentError(RuntimeError):
    """An error with the runtime environment."""


class MissingDependencyError(EnvironmentError):
    """A necessary dependency is not available."""


class EntityExistsError(RuntimeError):
    pass


class EntityNotFoundError(RuntimeError):
    pass


class ConcurrentModificationError(RuntimeError):
    pass


class UnsatisfiedPreconditionError(RuntimeError):
    pass



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import functools
import hashlib
import io
import json

# mypy: disable-error-code="import-untyped"
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any, BinaryIO, Iterable, Optional, Tuple
from urllib.parse import urlparse

import minio
import minio.api
import minio.helpers
import minio.time

from dyff.schema.dataset import max_artifact_size_bytes
from dyff.schema.errors import PlatformError
from dyff.schema.platform import Artifact, ArtifactURL, Entities, StorageSignedURL
from dyff.storage import paths
from dyff.storage.backend.base.storage import StorageBackend
from dyff.storage.config import config


class _MyMinio(minio.Minio):
    def presigned_url_for_base(
        self,
        base_url: str,
        method: str,
        bucket_name: str,
        object_name: str,
        expires: timedelta = timedelta(days=7),
        response_headers: minio.helpers.DictType | None = None,
        request_date: datetime | None = None,
        version_id: str | None = None,
        extra_query_params: minio.helpers.DictType | None = None,
    ) -> str:
        """Get presigned URL of an object relative to an arbitrary base URL, for HTTP
        method, expiry time and custom request parameters.

        The base URL can contain a "netloc" and an optional scheme. If a scheme
        is provided, it must be either 'http' or 'https'.

        :param base_url: The base URL where the client will access the s3
            system. Examples: ``s3.some.domain:9000``, ``https://other.domain``.
        :param method: HTTP method.
        :param bucket_name: Name of the bucket.
        :param object_name: Object name in the bucket.
        :param expires: Expiry in seconds; defaults to 7 days.
        :param response_headers: Optional response_headers argument to
                                 specify response fields like date, size,
                                 type of file, data about server, etc.
        :param request_date: Optional request_date argument to
                             specify a different request date. Default is
                             current date.
        :param version_id: Version ID of the object.
        :param extra_query_params: Extra query parameters for advanced usage.
        :return: URL string.

        Example::
            # Get presigned URL string to PUT 'my-object' in
            # 'my-bucket' with one day expiry, where the client
            # will access the s3 system at 'https://s3.mysite.com'.
            url = client.presigned_url_for_base(
                "https://s3.mysite.com",
                "PUT",
                "my-bucket",
                "my-object",
                expires=timedelta(days=1),
            )
            print(url)
        """
        # Same as get_presigned_url(), except we modify the URL before signing

        minio.api.check_bucket_name(bucket_name, s3_check=self._base_url.is_aws_host)
        minio.api.check_non_empty_string(object_name)
        if expires.total_seconds() < 1 or expires.total_seconds() > 604800:
            raise ValueError("expires must be between 1 second to 7 days")

        region = self._get_region(bucket_name)
        query_params = extra_query_params or {}
        query_params.update({"versionId": version_id} if version_id else {})
        query_params.update(response_headers or {})
        creds = self._provider.retrieve() if self._provider else None
        if creds and creds.session_token:
            query_params["X-Amz-Security-Token"] = creds.session_token
        url = self._base_url.build(
            method,
            region,
            bucket_name=bucket_name,
            object_name=object_name,
            query_params=query_params,
        )

        # Modify the URL
        parsed_base_url = urlparse(base_url)
        if not parsed_base_url.netloc:
            parsed_base_url = urlparse(f"https://{base_url}")
        if any(
            [
                parsed_base_url.path,
                parsed_base_url.query,
                parsed_base_url.params,
                parsed_base_url.fragment,
            ]
        ):
            raise ValueError(
                f"'base_url' must be only 'scheme://netloc' ; got {base_url}"
            )
        url = minio.helpers.url_replace(
            url, scheme=parsed_base_url.scheme, netloc=parsed_base_url.netloc
        )

        if creds:
            url = minio.api.presign_v4(
                method,
                url,
                region,
                creds,
                request_date or minio.time.utcnow(),
                int(expires.total_seconds()),
            )
        return minio.api.urlunsplit(url)


def _bucket_name_from_path(path: str) -> str:
    return _split_bucket_path(path)[0]


def _drop_protocol(path: str) -> str:
    protocol = "s3://"
    if path.startswith(protocol):
        path = path[len(protocol) :]
    return path


def _split_bucket_path(path: str) -> Tuple[str, str]:
    path = _drop_protocol(path)
    parts = path.split("/")
    return parts[0], "/".join(parts[1:])


def _configured_internal_endpoint() -> str:
    if config.storage.s3.internal_endpoint is not None:
        return config.storage.s3.internal_endpoint
    else:
        return config.storage.s3.endpoint


def _get_client() -> _MyMinio:
    endpoint = _configured_internal_endpoint()
    parsed_endpoint = urlparse(endpoint, scheme="https")
    if parsed_endpoint.scheme == "http":
        secure = False
    else:
        secure = True
    parsed_endpoint = parsed_endpoint._replace(scheme="")
    return _MyMinio(
        # Minio doesn't allow paths in the URL
        endpoint=parsed_endpoint.netloc,
        access_key=config.storage.s3.access_key,
        secret_key=config.storage.s3.secret_key.get_secret_value(),
        secure=secure,
    )


def _translate_errors(fn):
    @functools.wraps(fn)
    def _impl(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except minio.S3Error as ex:
            if ex.code in ["NoSuchKey", "NoSuchBucket", "ResourceNotFound"]:
                raise KeyError(f"{ex._resource} ({ex.code})")
            raise

    return _impl


# The following function is adapted from the following project:
# https://github.com/p2p-ld/numpydantic/blob/66fffc49f87bfaaa2f4d05bf1730c343b10c9cc6/src/numpydantic/serialization.py#L107-L142
# MIT License: https://github.com/p2p-ld/numpydantic/blob/main/LICENSE
def _py312_relative_path(
    self: Path | str, other: Path | str, walk_up: bool = True
) -> Path:
    """ "Backport" of :meth:`pathlib.Path.relative_to` with ``walk_up=True`` that's not
    available pre 3.12.

    Return the relative path to another path identified by the passed
    arguments.  If the operation is not possible (because this is not
    related to the other path), raise ValueError.

    The *walk_up* parameter controls whether `..` may be used to resolve
    the path.

    References:
        https://github.com/python/cpython/blob/8a2baedc4bcb606da937e4e066b4b3a18961cace/Lib/pathlib/_abc.py#L244-L270
    """
    if not isinstance(self, Path):
        self = Path(self)
    if not isinstance(other, Path):
        other = Path(other)
    self_parts = self.parts
    other_parts = other.parts
    anchor0, parts0 = self_parts[0], list(reversed(self_parts[1:]))
    anchor1, parts1 = other_parts[0], list(reversed(other_parts[1:]))
    if anchor0 != anchor1:
        raise ValueError(f"{self!r} and {other!r} have different anchors")
    while parts0 and parts1 and parts0[-1] == parts1[-1]:
        parts0.pop()
        parts1.pop()
    for part in parts1:
        if not part or part == ".":
            pass
        elif not walk_up:
            raise ValueError(f"{self!r} is not in the subpath of {other!r}")
        elif part == "..":
            raise ValueError(f"'..' segment in {other!r} cannot be walked")
        else:
            parts0.append("..")
    return Path(*reversed(parts0))


class S3StorageBackend(StorageBackend):
    @_translate_errors
    def storage_size(self, path: str) -> int:
        bucket, prefix = _split_bucket_path(path)
        client = _get_client()
        objects = client.list_objects(bucket, prefix, recursive=True)
        return sum(obj.size for obj in objects)

    @_translate_errors
    def list_dir(self, path: str, *, recursive: bool = False) -> Iterable[str]:
        if not path.startswith("s3://"):
            raise ValueError("path must be an s3 object")
        if not path.endswith("/"):
            path += "/"
        client = _get_client()
        remote_path = path[len("s3://") :]
        bucket_name, *rest = remote_path.split("/")
        prefix = "/".join(rest)
        objects = client.list_objects(bucket_name, prefix, recursive=recursive)
        yield from (f"s3://{obj.bucket_name}/{obj.object_name}" for obj in objects)

    @_translate_errors
    def download_recursive(self, source: str, destination: str) -> None:
        if not source.startswith("s3://"):
            raise ValueError("source must be an s3 object")
        if not source.endswith("/"):
            source += "/"
        client = _get_client()
        remote_path = source[len("s3://") :]
        bucket_name, *rest = remote_path.split("/")
        prefix = "/".join(rest)
        objects = client.list_objects(bucket_name, prefix, recursive=True)

        destination_resolved = Path(destination).resolve()

        for obj in objects:
            if obj.object_name.endswith("/"):
                # Not a file
                continue
            remote_file = Path(obj.object_name).relative_to(prefix)
            remote_directory = Path(obj.object_name).parent.relative_to(prefix)
            local_directory = destination_resolved / remote_directory
            local_directory.mkdir(parents=True, exist_ok=True)
            local_file = destination_resolved / remote_file
            client.fget_object(obj.bucket_name, obj.object_name, str(local_file))

        def require_tree_contains_child(tree: Path, child: Path) -> None:
            try:
                (tree / child).resolve().relative_to(tree)
            except ValueError as ex:
                raise ValueError(f"{child} is not a descendent of {tree}") from ex

        # Restore symlinks, if present
        # Note that this only looks for the symlinks.json file in the root
        # directory; it's meant to restore a matching call to upload_recursive()
        try:
            with open(destination_resolved / ".dyff" / "symlinks.json", "r") as fin:
                symlinks: list[dict[str, Any]] = json.load(fin)["symlinks"]
        except FileNotFoundError:
            pass
        else:
            print(f"restoring symlinks:\n{json.dumps(symlinks, indent=4)}")
            for entry in symlinks:
                link = destination_resolved / str(entry["link"])
                require_tree_contains_child(destination_resolved, link)
                target = Path(entry["target"])
                target_is_directory = bool(entry["target_is_directory"])
                if target.is_absolute():
                    # Not allowed because things will break if an absolute path
                    # is restored with a different prefix.
                    raise ValueError(
                        f"absolute symlink targets are not allowed: {target}"
                    )
                # The link target is relative to the link file
                # Note: This will follow links-to-links. It will not catch the
                # case where a link points outside the tree, but it points to
                # another link that points back inside the tree.
                require_tree_contains_child(destination_resolved, link / target)
                print(f"link: {link} -> {target}")
                link.parent.mkdir(exist_ok=True, parents=True)
                link.symlink_to(target, target_is_directory=target_is_directory)

    @_translate_errors
    def upload_recursive(self, source: str, destination: str) -> None:
        if not destination.startswith("s3://"):
            raise ValueError("destination must be an s3 object")
        if not destination.endswith("/"):
            destination += "/"

        source_path = Path(source).resolve()
        if not source_path.is_dir():
            raise ValueError("source must be a local directory")

        client = _get_client()
        bucket_name, bucket_path = _split_bucket_path(destination)

        # We need to record symlink info out-of-band because s3 has no concept
        # of symlinks and it will upload the target file instead
        source_paths: list[Path] = []
        symlinks: list[dict[str, Any]] = []
        for path in source_path.rglob("*"):
            if path.is_file():
                if path.is_symlink():
                    target = path.readlink()
                    if target.is_absolute():
                        # The target has to be relative to the link file or
                        # restoration will fail if the path prefix is different
                        target = _py312_relative_path(target, path.parent, walk_up=True)
                    symlinks.append(
                        {
                            "link": str(path.relative_to(source_path)),
                            "target": str(target),
                            "target_is_directory": target.is_dir(),
                        }
                    )
                else:
                    source_paths.append(path)
        relative_paths = [path.relative_to(source) for path in source_paths]

        if bucket_path != "":
            assert bucket_path[-1] == "/"  # See above
            prefix = bucket_path
        else:
            prefix = ""

        for file_path, upload_location in zip(source_paths, relative_paths):
            client.fput_object(
                bucket_name, f"{prefix}{upload_location}", str(file_path)
            )
        if symlinks:
            data = json.dumps({"symlinks": symlinks}).encode()
            client.put_object(
                bucket_name,
                f"{prefix}.dyff/symlinks.json",
                io.BytesIO(data),
                len(data),
            )

    @_translate_errors
    def put_object(self, data: bytes | BinaryIO, destination: str) -> None:
        if not destination.startswith("s3://"):
            raise ValueError("destination must be an s3 object")
        bucket_name, bucket_path = _split_bucket_path(destination)
        if bucket_path.endswith("/"):
            raise ValueError("destination is a directory")
        if isinstance(data, bytes):
            data = io.BytesIO(data)
        _get_client().put_object(
            bucket_name, bucket_path, data, length=-1, part_size=10 * 1024 * 1024
        )

    @_translate_errors
    def get_object(self, source: str) -> BinaryIO:
        if not source.startswith("s3://"):
            raise ValueError("source must be an s3 object")
        bucket_name, bucket_path = _split_bucket_path(source)
        if bucket_path.endswith("/"):
            raise ValueError("source is a directory")

        try:
            response = _get_client().get_object(bucket_name, bucket_path)
            return io.BytesIO(response.data)
        finally:
            response.close()
            response.release_conn()

    @_translate_errors
    def delete_object(self, destination: str) -> None:
        if not destination.startswith("s3://"):
            raise ValueError("destination must be an s3 object")
        bucket_name, bucket_path = _split_bucket_path(destination)
        if bucket_path.endswith("/"):
            raise ValueError("destination is a directory")

        _get_client().remove_object(bucket_name, bucket_path)

    @_translate_errors
    def delete_tree(self, destination: str) -> None:
        if not destination.startswith("s3://"):
            raise ValueError("destination must be an s3 object")
        bucket_name, bucket_path = _split_bucket_path(destination)
        if bucket_path.endswith("/"):
            raise ValueError("destination is a directory")

        # From code example here:
        # https://min.io/docs/minio/linux/developers/python/API.html#remove-objects-bucket-name-delete-object-list-bypass-governance-mode-false
        client = _get_client()
        to_delete = [
            minio.api.DeleteObject(obj.object_name)
            for obj in client.list_objects(bucket_name, bucket_path, recursive=True)
        ]
        errors = client.remove_objects(bucket_name, to_delete)
        if errors:
            raise PlatformError(f"Failed to remove objects: {errors}")

    @_translate_errors
    def signed_url_for_artifact_upload(
        self,
        artifact: Artifact,
        storage_path: str,
        *,
        size_limit_bytes: Optional[int] = None,
        _internal_client: bool = False,
    ) -> StorageSignedURL:
        """Create a temporary signed URL that can be used in a PUT request to upload an
        ``Artifact`` directly to storage."""
        if size_limit_bytes is None:
            size_limit_bytes = max_artifact_size_bytes()
        if artifact.digest.md5 is None:
            raise ValueError("requires artifact.digest.md5")
        bucket_name, bucket_path = _split_bucket_path(storage_path)
        client = _get_client()

        # If the s3 provider is self-hosted (e.g. Minio), external clients
        # connect to it at a different URL than internal clients. We must sign
        # the correct URL, depending on which kind of client will use it.
        if _internal_client and config.storage.s3.internal_endpoint is not None:
            configured_url = config.storage.s3.internal_endpoint
        else:
            configured_url = config.storage.s3.endpoint

        signed_url = client.presigned_url_for_base(
            configured_url,
            "PUT",
            bucket_name,
            f"{bucket_path}/{artifact.path}",
            expires=timedelta(hours=1),
        )

        headers: dict[str, str] = {
            # TODO: Is there an s3 equivalent of this?
            # Google custom header limiting the size of the artifact
            # "x-goog-content-length-range": f"0,{size_limit_bytes}",
        }
        return StorageSignedURL(url=signed_url, method="PUT", headers=headers)

    @_translate_errors
    def signed_url_for_dataset_upload(
        self,
        dataset_id: str,
        artifact: Artifact,
        *,
        size_limit_bytes: Optional[int] = None,
        storage_path: Optional[str] = None,
        _internal_client: bool = False,
    ) -> StorageSignedURL:
        storage_path = storage_path or paths.dataset_root(dataset_id)
        return self.signed_url_for_artifact_upload(
            artifact,
            storage_path,
            size_limit_bytes=size_limit_bytes,
            _internal_client=_internal_client,
        )

    def _artifact_url_for_downlink(
        self,
        *,
        root_path: str,
        artifact_path: str,
        client: _MyMinio | None = None,
        _internal_client: bool = False,
    ) -> ArtifactURL:
        bucket_name, bucket_path = _split_bucket_path(artifact_path)
        relative_path = Path(bucket_path).relative_to(root_path)
        # TODO: Populate contentType and digest
        artifact = Artifact(path=str(relative_path))

        if client is None:
            client = _get_client()

        # If the s3 provider is self-hosted (e.g. Minio), external clients
        # connect to it at a different URL than internal clients. We must sign
        # the correct URL, depending on which kind of client will use it.
        if _internal_client and config.storage.s3.internal_endpoint is not None:
            configured_url = config.storage.s3.internal_endpoint
        else:
            configured_url = config.storage.s3.endpoint

        signed_url = client.presigned_url_for_base(
            configured_url,
            "GET",
            bucket_name,
            bucket_path,
            expires=timedelta(hours=1),
        )

        headers: dict[str, str] = {}
        return ArtifactURL(
            artifact=artifact,
            signedURL=StorageSignedURL(url=signed_url, method="GET", headers=headers),
        )

    @_translate_errors
    def logs_downlink(
        self, entity_kind: Entities, entity_id: str, *, _internal_client: bool = False
    ) -> Optional[ArtifactURL]:
        if entity_kind not in [
            Entities.Measurement,
            Entities.Report,
            Entities.SafetyCase,
        ]:
            return None

        storage_root = paths.entity_artifacts_root(entity_kind, entity_id)
        _root_bucket, root_path = _split_bucket_path(storage_root)
        artifact_path = paths.logs_file_for_entity(storage_root)
        return self._artifact_url_for_downlink(
            root_path=root_path,
            artifact_path=artifact_path,
            _internal_client=_internal_client,
        )

    @_translate_errors
    def artifact_downlinks(
        self, entity_kind: Entities, entity_id: str, *, _internal_client: bool = False
    ) -> Iterable[ArtifactURL]:
        storage_root = paths.entity_artifacts_root(entity_kind, entity_id)
        _root_bucket, root_path = _split_bucket_path(storage_root)
        artifact_paths = self.list_dir(storage_root, recursive=True)
        client = _get_client()

        for artifact_path in artifact_paths:
            yield self._artifact_url_for_downlink(
                root_path=root_path,
                artifact_path=artifact_path,
                client=client,
                _internal_client=_internal_client,
            )

    @_translate_errors
    def object_md5hash(self, storage_path: str) -> bytes:
        bucket_name, bucket_path = _split_bucket_path(storage_path)
        client = _get_client()

        response = None
        try:
            response = client.get_object(bucket_name, bucket_path)
            if response is None:
                raise ValueError(f"no object {storage_path}")
            # FIXME: This is potentially a very expensive operation. We have to
            # compute it on the server for security, but we should cache it in
            # the object metadata if we can confirm that users can't alter the
            # metadata using the signed upload URLs.
            # Also consider the s3 mechanisms for pre-computing digests as
            # metadata (but note that it doesn't support md5):
            # https://aws.amazon.com/blogs/aws/new-additional-checksum-algorithms-for-amazon-s3/
            return hashlib.md5(response.read()).digest()
        finally:
            if response is not None:
                response.close()
                response.release_conn()

    @_translate_errors
    def dataset_artifact_md5hash(
        self, dataset_id: str, artifact_path: str, *, storage_path: Optional[str] = None
    ) -> bytes:
        storage_path = storage_path or paths.dataset_root(dataset_id)
        return self.object_md5hash(f"{storage_path}/{dataset_id}/{artifact_path}")


__all__ = ["S3StorageBackend"]



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import abc
from typing import Optional

from dyff.schema.platform import Account, APIKey, Identity


class AuthBackend(abc.ABC):
    @abc.abstractmethod
    def create_account(self, name: str, identity: Optional[Identity] = None) -> Account:
        """Create a new Account.

        Parameters:
        name: A unique human-readable name

        Returns:
        Entity representing the Account.
        """

    @abc.abstractmethod
    def delete_account(self, account_id: str):
        """Delete an Account.

        Parameters:
        account_id: The unique identifier of the Account.
        """

    @abc.abstractmethod
    def get_account(
        self,
        *,
        id: Optional[str] = None,
        name: Optional[str] = None,
        identity: Optional[Identity] = None,
    ) -> Optional[Account]:
        """Get an Account by ID, identity, or name.

        Parameters:
        id: The unique identifier of the Account.
        name: The unique name of the Account.
        identity: Unique identifier of the Account according to a third-party
          identity provider. It is an error to set more than one identity.
        """

    @abc.abstractmethod
    def add_api_key(self, account_id: str, api_key: APIKey) -> None:
        """Add a new APIKey to an Account.

        Parameters:
        account_id: The unique identifier of the Account.
        api_key: The new API key.
        """

    @abc.abstractmethod
    def revoke_api_key(self, account_id: str, api_key_id: str) -> None:
        """Revoke an APIKey associated with an Account.

        Parameters:
        account_id: The unique identifier of the Account.
        api_key: The unique identifier of the APIKey.
        """

    @abc.abstractmethod
    def revoke_all_api_keys(self, account_id: str) -> None:
        """Revoke all API keys for the given Account.

        Parameters:
        account_id: The unique identifier of the Account.
        """

    @abc.abstractmethod
    def get_api_key(self, account_id: str, api_key_id: str) -> Optional[APIKey]:
        """Get an APIKey associated with an Account by ID.

        Parameters:
        account_id: The unique identifier of the Account.
        api_key_id: The unique identifier of the APIKey.
        """

    @abc.abstractmethod
    def get_all_api_keys(self, account_id: str) -> list[APIKey]:
        """Get the API keys associated with an Account.

        Parameters:
        account_id: The unique identifier of the Account.
        """



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import json
from datetime import datetime, timezone
from typing import Collection, Iterable, Optional, TypeVar

import pymongo

from dyff.schema.errors import ClientError
from dyff.schema.platform import (
    Audit,
    Dataset,
    DataSource,
    Documentation,
    DyffEntity,
    Entities,
    Evaluation,
    Family,
    InferenceService,
    InferenceSession,
    Labeled,
    Measurement,
    Method,
    Model,
    Module,
    Report,
    Resources,
    SafetyCase,
    Score,
    UseCase,
)
from dyff.schema.requests import (
    AuditQueryRequest,
    DatasetQueryRequest,
    DocumentationEditRequest,
    DocumentationQueryRequest,
    EvaluationQueryRequest,
    FamilyQueryRequest,
    InferenceServiceQueryRequest,
    InferenceSessionQueryRequest,
    MeasurementQueryRequest,
    MethodQueryRequest,
    ModelQueryRequest,
    ModuleQueryRequest,
    ReportQueryRequest,
    SafetyCaseQueryRequest,
    ScoreQueryRequest,
    UseCaseQueryRequest,
)
from dyff.storage.backend.base.query import QueryBackend, Whitelist
from dyff.storage.config import config

_DyffEntityT = TypeVar("_DyffEntityT", bound=DyffEntity)


class MongoDBQueryBackend(QueryBackend):
    # Mapping of query keys in the API to the corresponding DB fields
    query_key_map = {
        Entities.Evaluation: {
            "inferenceService": "inferenceSession.inferenceService.id",
            "inferenceServiceName": "inferenceSession.inferenceService.name",
            "model": "inferenceSession.inferenceService.model.id",
            "modelName": "inferenceSession.inferenceService.model.name",
        },
        Entities.InferenceService: {
            "model": "model.id",
            "modelName": "model.name",
        },
        Entities.InferenceSession: {
            "inferenceService": "inferenceService.id",
            "inferenceServiceName": "inferenceService.name",
            "model": "inferenceService.model.id",
            "modelName": "inferenceService.model.name",
        },
        Entities.Measurement: {
            "method": "method.id",
            "methodName": "method.name",
            "dataset": "scope.dataset",
            "evaluation": "scope.evaluation",
            "inferenceService": "scope.inferenceService",
            "model": "scope.model",
        },
        Entities.Method: {
            "outputKind": "output.kind",
        },
        Entities.SafetyCase: {
            "method": "method.id",
            "methodName": "method.name",
            "dataset": "scope.dataset",
            "evaluation": "scope.evaluation",
            "inferenceService": "scope.inferenceService",
            "model": "scope.model",
        },
        Entities.Score: {
            "method": "metadata.refs.method",
            "dataset": "metadata.refs.dataset",
            "evaluation": "metadata.refs.evaluation",
            "inferenceService": "metadata.refs.inferenceService",
            "model": "metadata.refs.model",
        },
    }

    def __init__(self):
        connection_string = config.api.query.mongodb.connection_string
        self._client = pymongo.MongoClient(connection_string.get_secret_value())
        self._workflows_db = self._client.get_database(
            config.api.query.mongodb.database
        )

    def _convert_entity_response(self, kind: Entities, entity: dict) -> dict:
        def array_to_object(xs: list[dict]) -> dict:
            return {x["key"]: x["value"] for x in xs}

        if kind == Entities.Documentation:
            entity = dict(entity)
            entity["entity"] = entity["_id"]
            del entity["_id"]
            return entity
        else:
            entity = dict(entity)
            entity["id"] = entity["_id"]

            # dicts with user-supplied keys are stored as [(k, v)] lists
            # and need to be converted back
            if (labels := entity.get("labels")) is not None:
                entity["labels"] = array_to_object(labels)
            # if (annotations := entity.get("annotations")) is not None:
            #     entity["annotations"] = array_to_object(annotations)
            if kind == Entities.Family:
                if (members := entity.get("members")) is not None:
                    entity["members"] = array_to_object(members)

            del entity["_id"]
            return entity

    def _filter_from_whitelist(self, whitelist: Whitelist) -> Optional[dict]:
        if ("*" not in whitelist.accounts) and ("*" not in whitelist.entities):
            # Query constraint requiring the result to be in the whitelist
            return {
                "$or": [
                    {"account": {"$in": list(whitelist.accounts)}},
                    {"_id": {"$in": list(whitelist.entities)}},
                ]
            }
        else:
            return None

    def _get_entity(self, kind: Entities, id: str) -> Optional[dict]:
        collection_name = Resources.for_kind(kind)
        collection = self._workflows_db[collection_name]
        result = collection.find_one({"_id": id})
        if result:
            result = self._convert_entity_response(kind, result)
        return result

    def _query_entities(
        self, kind: Entities, whitelist: Whitelist, query: dict
    ) -> Iterable[dict]:
        print(f"whitelist: {whitelist}")
        mongo_conjunction: list[dict] = []
        if whitelist_filter := self._filter_from_whitelist(whitelist):
            mongo_conjunction.append(whitelist_filter)

        def json_decode(k: str, v: str):
            try:
                return json.loads(v)
            except json.decoder.JSONDecodeError:
                raise ClientError(f"{k}: json decode failed")

        def build_query(query: dict, key_map: dict[str, str]) -> list:
            conjunction: list = []
            for k, v in query.items():
                if v is None:
                    continue

                if k == "id":
                    k = "_id"
                else:
                    # Map query key to DB key, if mapping exists
                    k = key_map.get(k, k)

                if k == "labels":
                    if isinstance(v, dict):
                        labels = v
                    else:
                        labels = json_decode(k, v)
                    labeled = Labeled(labels=labels)  # validate
                    for label_key, label_value in labeled.labels.items():
                        conjunction.append(
                            {
                                k: {
                                    "$elemMatch": {
                                        "key": label_key,
                                        "value": label_value,
                                    }
                                }
                            }
                        )
                # TODO:
                # elif k == "annotations":
                #     pass
                elif k == "inputs":
                    # This is a special query for Analysis-related entities. It
                    # selects entities that reference any of the listed IDs in
                    # their .inputs field.
                    if isinstance(v, list):
                        entity_list = v
                    elif isinstance(v, str):
                        entity_list = v.split(",")
                    else:
                        raise ClientError(f"inputs: must be list or str; got {type(v)}")
                    conjunction.append(
                        {"inputs": {"$elemMatch": {"entity": {"$in": entity_list}}}}
                    )
                elif isinstance(v, list):
                    conjunction.append({k: {"$in": list(v)}})
                else:
                    conjunction.append({k: v})
            return conjunction

        key_map = MongoDBQueryBackend.query_key_map.get(kind, {})

        # These are options to the query operation and need to be removed
        # from the actual query
        options = {}
        if orderBy := query.pop("orderBy", None):
            order = query.pop("order", None)
            options["sort"] = {
                orderBy: (
                    pymongo.DESCENDING if order == "descending" else pymongo.ASCENDING
                ),
            }
        if limit := query.pop("limit", None):
            options["limit"] = limit

        json_query_string = query.pop("query", None)
        if json_query_string is not None:
            json_query_object = json_decode("query", json_query_string)
            if isinstance(json_query_object, list):
                for item in json_query_object:
                    mongo_conjunction.extend(build_query(item, key_map))
            elif isinstance(json_query_object, dict):
                mongo_conjunction.extend(build_query(json_query_object, key_map))
            else:
                raise ClientError(
                    f"query: must parse as dict or list[dict]; got {type(json_query_object)}"
                )

        mongo_conjunction.extend(build_query(query, key_map))

        mongo_query = {"$and": mongo_conjunction} if mongo_conjunction else {}
        print(f"query: {mongo_query}")
        collection_name = Resources.for_kind(kind)
        collection = self._workflows_db[collection_name]

        results = collection.find(mongo_query, **options)

        for result in results:
            yield self._convert_entity_response(kind, result)

    def expired_entities(
        self, entity_type: type[_DyffEntityT], *, before: datetime
    ) -> list[_DyffEntityT]:
        before = before.astimezone(timezone.utc)
        kind = Entities(entity_type.__name__)
        collection_name = Resources.for_kind(kind)
        collection = self._workflows_db[collection_name]
        query = {
            "status": "Deleted",
            "lastTransitionTime": {"$ne": None, "$lt": before.isoformat()},
        }
        results = collection.find(query)
        return [
            entity_type.parse_obj(self._convert_entity_response(kind, result))
            for result in results
        ]

    # ------------------------------------------------------------------------

    def get_documentation(self, resource_id: str) -> Optional[Documentation]:
        """Retrieve the Documentation entity associated with a resource.

        :param resource_id: ID of the documented resource.
        :returns: The Documentation, or None if no resource exists with the given ID.
        """
        result = self._get_entity(Entities.Documentation, resource_id)
        return Documentation.parse_obj(result) if result else None

    def query_documentation(
        self, whitelist: Whitelist, query: DocumentationQueryRequest
    ) -> Collection[Documentation]:
        """Retrieve all Documentation entities matching the query parameters.

        :param whitelist: The set of accounts and entities that the caller has been
            granted access to.
        :param query: The query request.
        """
        results = self._query_entities(Entities.Documentation, whitelist, query.dict())
        return [Documentation.parse_obj(result) for result in results]

    def get_audit(self, id: str) -> Optional[Audit]:
        """Retrieve an Audit entity.

        :param id: The unique key of the Audit.
        :returns: The Audit, or None if no Audit with the specified key exists.
        """
        result = self._get_entity(Entities.Audit, id)
        return Audit.parse_obj(result) if result else None

    def query_audits(
        self, whitelist: Whitelist, query: AuditQueryRequest
    ) -> Collection[Audit]:
        """Retrieve all Audit entities matching the query parameters.

        :param whitelist: The set of accounts and entities that the caller has been
            granted access to.
        :param query: Equality constraints on fields of the Audit entity. The returned
            entities satisfy 'entity.field==value' for all items 'field: value' in
            kwargs.
        """
        results = self._query_entities(Entities.Audit, whitelist, query.dict())
        return [Audit.parse_obj(result) for result in results]

    def get_data_source(self, id: str) -> Optional[DataSource]:
        """Retrieve a DataSource entity.

        :param id: The unique key of the DataSource.
        :returns: The DataSource, or None if no DataSource with the specified key
            exists.
        """
        result = self._get_entity(Entities.DataSource, id)
        return DataSource.parse_obj(result) if result else None

    def query_data_sources(
        self, whitelist: Whitelist, **query
    ) -> Collection[DataSource]:
        """Retrieve all DataSource entities matching the query parameters.

        :param whitelist: The set of accounts and entities that the caller has been
            granted access to.
        :param query: Equality constraints on fields of the DataSource entity. The
            returned entities satisfy 'entity.field==value' for all items 'field: value'
            in kwargs.
        """
        results = self._query_entities(Entities.DataSource, whitelist, query)
        return [DataSource.parse_obj(result) for result in results]

    def get_dataset(self, id: str) -> Optional[Dataset]:
        """Retrieve a Dataset entity.

        :param id: The unique key of the Dataset.
        :returns: The Dataset, or None if no Dataset with the specified key exists.
        """
        result = self._get_entity(Entities.Dataset, id)
        return Dataset.parse_obj(result) if result else None

    def query_datasets(
        self, whitelist: Whitelist, query: DatasetQueryRequest
    ) -> Collection[Dataset]:
        """Retrieve all Dataset entities matching the query parameters.

        :param whitelist: The set of accounts and entities that the caller has been
            granted access to.
        :param query: Equality constraints on fields of the Dataset entity. The returned
            entities satisfy 'entity.field==value' for all items 'field: value' in
            kwargs.
        """
        results = self._query_entities(Entities.Dataset, whitelist, query.dict())
        return [Dataset.parse_obj(result) for result in results]

    def get_evaluation(self, id: str) -> Optional[Evaluation]:
        """Retrieve an Evaluation entity.

        :param id: The unique key of the Evaluation.
        :returns: The Evaluation, or None if no Evaluation with the specified key
            exists.
        """
        result = self._get_entity(Entities.Evaluation, id)
        return Evaluation.parse_obj(result) if result else None

    def query_evaluations(
        self, whitelist: Whitelist, query: EvaluationQueryRequest
    ) -> Collection[Evaluation]:
        """Retrieve all Evaluation entities matching the query parameters.

        :param whitelist: The set of accounts and entities that the caller has been
            granted access to.
        :param query: Equality constraints on fields of the Evaluation entity. The
            returned entities satisfy 'entity.field==value' for all items 'field: value'
            in kwargs.
        """
        results = self._query_entities(Entities.Evaluation, whitelist, query.dict())
        return [Evaluation.parse_obj(result) for result in results]

    def get_family(self, id: str) -> Optional[Family]:
        """Retrieve a Family entity.

        :param id: The unique ID of the Family.
        :returns: The Family, or None if no Family with the specified ID exists.
        """
        result = self._get_entity(Entities.Family, id)
        return Family.parse_obj(result) if result else None

    def query_families(
        self, whitelist: Whitelist, query: FamilyQueryRequest
    ) -> Collection[Family]:
        """Retrieve all Family entities matching the query parameters.

        :param whitelist: The set of accounts and entities that the caller has been
            granted access to.
        :param query: Equality constraints on fields of the Family entity. The returned
            entities satisfy 'entity.field==value' for all items 'field: value' in
            kwargs.
        """
        results = self._query_entities(Entities.Family, whitelist, query.dict())
        return [Family.parse_obj(result) for result in results]

    def get_inference_service(self, id: str) -> Optional[InferenceService]:
        """Retrieve an InferenceService entity.

        :param id: The unique key of the InferenceService.
        :returns: The InferenceService, or None if no InferenceService with the
            specified key exists.
        """
        result = self._get_entity(Entities.InferenceService, id)
        return InferenceService.parse_obj(result) if result else None

    def query_inference_services(
        self, whitelist: Whitelist, query: InferenceServiceQueryRequest
    ) -> Collection[InferenceService]:
        """Retrieve all InferenceService entities matching the query parameters.

        :param whitelist: The set of accounts and entities that the caller has been
            granted access to.
        :param query: Equality constraints on fields of the InferenceService entity. The
            returned entities satisfy 'entity.field==value' for all items 'field: value'
            in kwargs.
        """
        results = self._query_entities(
            Entities.InferenceService, whitelist, query.dict()
        )
        return [InferenceService.parse_obj(result) for result in results]

    def get_inference_session(self, id: str) -> Optional[InferenceSession]:
        """Retrieve an InferenceSession entity.

        :param id: The unique key of the InferenceSession.
        :returns: The InferenceSession, or None if no InferenceSession with the
            specified key exists.
        """
        result = self._get_entity(Entities.InferenceSession, id)
        return InferenceSession.parse_obj(result) if result else None

    def query_inference_sessions(
        self, whitelist: Whitelist, query: InferenceSessionQueryRequest
    ) -> Collection[InferenceSession]:
        """Retrieve all InferenceSession entities matching the query parameters.

        :param whitelist: The set of accounts and entities that the caller has been
            granted access to.
        :param query: Equality constraints on fields of the InferenceSession entity. The
            returned entities satisfy 'entity.field==value' for all items 'field: value'
            in kwargs.
        """
        results = self._query_entities(
            Entities.InferenceSession, whitelist, query.dict()
        )
        return [InferenceSession.parse_obj(result) for result in results]

    def get_measurement(self, id: str) -> Optional[Measurement]:
        """Retrieve a Measurement entity.

        :param id: The unique key of the Measurement.
        :returns: The Measurement, or None if no Measurement with the specified key
            exists.
        """
        result = self._get_entity(Entities.Measurement, id)
        return Measurement.parse_obj(result) if result else None

    def query_measurements(
        self, whitelist: Whitelist, query: MeasurementQueryRequest
    ) -> Collection[Measurement]:
        """Retrieve all Measurement entities matching the query parameters.

        :param whitelist: The set of accounts and entities that the caller has been
            granted access to.
        :param query: Equality constraints on fields of the Measurement entity. The
            returned entities satisfy 'entity.field==value' for all items 'field: value'
            in kwargs.
        """
        results = self._query_entities(Entities.Measurement, whitelist, query.dict())
        return [Measurement.parse_obj(result) for result in results]

    def get_method(self, id: str) -> Optional[Method]:
        """Retrieve a Method entity.

        :param id: The unique key of the Method.
        :returns: The Method, or None if no Method with the specified key exists.
        """
        result = self._get_entity(Entities.Method, id)
        return Method.parse_obj(result) if result else None

    def query_methods(
        self, whitelist: Whitelist, query: MethodQueryRequest
    ) -> Collection[Method]:
        """Retrieve all Method entities matching the query parameters.

        :param whitelist: The set of accounts and entities that the caller has been
            granted access to.
        :param query: Equality constraints on fields of the Method entity. The returned
            entities satisfy 'entity.field==value' for all items 'field: value' in
            kwargs.
        """
        results = self._query_entities(Entities.Method, whitelist, query.dict())
        return [Method.parse_obj(result) for result in results]

    def get_model(self, id: str) -> Optional[Model]:
        """Retrieve a Model entity.

        :param id: The unique key of the Model.
        :returns: The Model, or None if no Model with the specified key exists.
        """
        result = self._get_entity(Entities.Model, id)
        return Model.parse_obj(result) if result else None

    def query_models(
        self, whitelist: Whitelist, query: ModelQueryRequest
    ) -> Collection[Model]:
        """Retrieve all Model entities matching the query parameters.

        :param whitelist: The set of accounts and entities that the caller has been
            granted access to.
        :param query: Equality constraints on fields of the Model entity. The returned
            entities satisfy 'entity.field==value' for all items 'field: value' in
            kwargs.
        """
        results = self._query_entities(Entities.Model, whitelist, query.dict())
        return [Model.parse_obj(result) for result in results]

    def get_module(self, id: str) -> Optional[Module]:
        """Retrieve a Module entity.

        :param id: The unique key of the Module.
        :returns: The Module, or None if no Module with the specified key exists.
        """
        result = self._get_entity(Entities.Module, id)
        return Module.parse_obj(result) if result else None

    def query_modules(
        self, whitelist: Whitelist, query: ModuleQueryRequest
    ) -> Collection[Module]:
        """Retrieve all Module entities matching the query parameters.

        :param whitelist: The set of accounts and entities that the caller has been
            granted access to.
        :param query: Equality constraints on fields of the Module entity. The returned
            entities satisfy 'entity.field==value' for all items 'field: value' in
            kwargs.
        """
        results = self._query_entities(Entities.Module, whitelist, query.dict())
        return [Module.parse_obj(result) for result in results]

    def get_report(self, id: str) -> Optional[Report]:
        """Retrieve a Report entity.

        :param id: The unique key of the Report.
        :returns: The Report, or None if no Report with the specified key exists.
        """
        result = self._get_entity(Entities.Report, id)
        return Report.parse_obj(result) if result else None

    def query_reports(
        self, whitelist: Whitelist, query: ReportQueryRequest
    ) -> Collection[Report]:
        """Retrieve all Report entities matching the query parameters.

        :param whitelist: The set of accounts and entities that the caller has been
            granted access to.
        :param query: Equality constraints on fields of the Report entity. The returned
            entities satisfy 'entity.field==value' for all items 'field: value' in
            kwargs.
        """
        results = self._query_entities(Entities.Report, whitelist, query.dict())
        return [Report.parse_obj(result) for result in results]

    def get_safetycase(self, id: str) -> Optional[SafetyCase]:
        """Retrieve a SafetyCase entity.

        :param id: The unique key of the SafetyCase.
        :returns: The SafetyCase, or None if no SafetyCase with the specified key
            exists.
        """
        result = self._get_entity(Entities.SafetyCase, id)
        return SafetyCase.parse_obj(result) if result else None

    def query_safetycases(
        self, whitelist: Whitelist, query: SafetyCaseQueryRequest
    ) -> Collection[SafetyCase]:
        """Retrieve all SafetyCase entities matching the query parameters.

        :param whitelist: The set of accounts and entities that the caller has been
            granted access to.
        :param query: Equality constraints on fields of the SafetyCase entity. The
            returned entities satisfy 'entity.field==value' for all items 'field: value'
            in kwargs.
        """
        results = self._query_entities(Entities.SafetyCase, whitelist, query.dict())
        return [SafetyCase.parse_obj(result) for result in results]

    def get_score(self, id: str) -> Optional[Score]:
        """Retrieve a Score entity.

        :param id: The unique key of the Score.
        :returns: The Score, or None if no Score with the specified key exists.
        """
        result = self._get_entity(Entities.Score, id)
        return Score.parse_obj(result) if result else None

    def query_scores(
        self, whitelist: Whitelist, query: ScoreQueryRequest
    ) -> Collection[Score]:
        """Retrieve all Score entities matching the query parameters.

        :param whitelist: The set of accounts and entities that the caller has been
            granted access to.
        :param query: Equality constraints on fields of the Score entity. The returned
            entities satisfy 'entity.field==value' for all items 'field: value' in
            kwargs.
        """
        results = self._query_entities(Entities.Score, whitelist, query.dict())
        return [Score.parse_obj(result) for result in results]

    def get_usecase(self, id: str) -> Optional[UseCase]:
        """Retrieve a UseCase entity.

        :param id: The unique key of the UseCase.
        :returns: The UseCase, or None if no UseCase with the specified key exists.
        """
        result = self._get_entity(Entities.UseCase, id)
        return UseCase.parse_obj(result) if result else None

    def query_usecases(
        self, whitelist: Whitelist, query: UseCaseQueryRequest
    ) -> Collection[UseCase]:
        """Retrieve all UseCase entities matching the query parameters.

        :param whitelist: The set of accounts and entities that the caller has been
            granted access to.
        :param query: Equality constraints on fields of the UseCase entity. The returned
            entities satisfy 'entity.field==value' for all items 'field: value' in
            kwargs.
        """
        results = self._query_entities(Entities.UseCase, whitelist, query.dict())
        return [UseCase.parse_obj(result) for result in results]

    # ------------------------------------------------------------------------
    # FIXME: Some of these methods don't belong on the "query" side, because
    # they mutate state. But, these "metadata" methods don't go through the
    # Kafka backend because the other backend components don't care about
    # metadata. Maybe there should be a third "metadata" backend to contain
    # these?

    def edit_documentation(
        self, id: str, edit: DocumentationEditRequest
    ) -> Optional[Documentation]:
        """Edit the Documentation entity associated with a resource.

        :param id: ID of the documented resource.
        :param edit: Edit request containing changes to make to the documentation.
        :returns: The new Documentation, or None if no resource exists with the given
            ID.
        """
        collection = self._workflows_db["documentation"]
        edit_dict = edit.documentation.dict()
        result = collection.find_one_and_update(
            {"_id": id},
            {"$set": edit_dict},
            upsert=True,
            return_document=pymongo.ReturnDocument.AFTER,
        )
        if result is None:
            return None
        del result["_id"]
        return Documentation.parse_obj(result)



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

# mypy: disable-error-code="import-untyped"
import json
from typing import TypeVar

import pydantic
from confluent_kafka import Producer

from dyff.schema.commands import DyffCommandType
from dyff.schema.platform import DyffEntity, DyffEntityType
from dyff.storage.config import config
from dyff.storage.typing import YAMLObject

from ..base.command import CommandBackend


def serialize_id(id: str) -> bytes:
    return id.encode("utf-8")


def serialize_value(value: YAMLObject) -> bytes:
    return json.dumps(value).encode("utf-8")


def serialize_entity(entity: DyffEntity) -> bytes:
    return entity.json().encode("utf-8")


def deserialize_id(id: bytes) -> str:
    return id.decode("utf-8")


def deserialize_value(value: bytes) -> YAMLObject:
    return json.loads(value.decode("utf-8"))


def deserialize_entity(entity: bytes) -> DyffEntityType:
    s = entity.decode("utf-8")
    # DyffEntityType is a pydantic "discriminated union", so pydantic infers
    # the type from the '.kind' field
    return pydantic.parse_raw_as(DyffEntityType, s)  # type: ignore


EntityT = TypeVar("EntityT", bound=DyffEntity)


class KafkaCommandBackend(CommandBackend):
    """Implementation of the Command model using Kafka streams."""

    def __init__(self):
        self.events_topic = config.kafka.topics.workflows_events
        producer_config = config.kafka.config.get_producer_config()
        # FIXME: This should be the global default. It's supposed to be the
        # default in Kafka > 3.0, but the docs of librdkafka suggest that
        # its default is still False.
        # See: https://github.com/confluentinc/librdkafka/blob/master/CONFIGURATION.md
        producer_config["enable.idempotence"] = True
        # FIXME: See comment in alignmentlabs.dyff.web.server
        # Can't get real logging to work when running in uvicorn
        print(
            f"Creating KafkaCommandBackend with config:\n{json.dumps(producer_config, indent=2)}",
            flush=True,
        )
        # logging.info(f"Creating KafkaCommandBackend with config:\n{json.dumps(producer_config, indent=2)}")
        self._kafka_producer = Producer(producer_config)

    def close(self) -> None:
        """Shut down the command backend cleanly."""
        if self._kafka_producer:
            self._kafka_producer.flush()
        self._kafka_producer = None

    def execute(self, command: DyffCommandType) -> None:
        message = command.model_dump(mode="json")
        self._kafka_producer.produce(
            topic=self.events_topic,
            value=serialize_value(message),
            # TODO: There may one day be commands that don't have a .data.id
            # property because they don't apply to a single entity.
            key=serialize_id(command.data.id),
        )



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import importlib


def symbol(fully_qualified_name):
    tokens = fully_qualified_name.split(".")
    module_name = ".".join(tokens[:-1])
    member = tokens[-1]
    module = importlib.import_module(module_name)
    return getattr(module, member)


def instantiate(fully_qualified_name, *args, **kwargs):
    constructor = symbol(fully_qualified_name)
    return constructor(*args, **kwargs)
